from __future__ import annotations

import warnings

import numpy as np
from sklearn.utils.validation import check_consistent_length, check_is_fitted
from tqdm.auto import tqdm

from .base import HubnessReduction


class CSLS(HubnessReduction):
    """Hubness reduction with Cross-domain similarity local scaling.

    Uses the formula presented in [1]_.

    Parameters
    ----------
    k: int, default = 5
        Number of neighbors to consider for mean distance of k-nearest neighbors
    verbose: int, default= 0
        Verbosity level

    References
    ----------
    .. [1] Lample, G., Conneau, A., Ranzato, M., Denoyer, L., & Jégou, H. (2018)
           Word translation without parallel data
           In: 6th International Conference on Learning Representations,
           ICLR 2018 - Conference Track Proceedings.
           https://openreview.net/forum?id=H196sainb
    """

    def __init__(self, k: int = 5, verbose: int = 0, *args, **kwargs):
        super().__init__(**kwargs)
        self.k = k
        self.verbose = verbose

    def __repr__(self):
        return f"{self.__class__.__name__}(k={self.k}, verbose = {self.verbose})"

    def fit(
        self,
        neigh_dist,
        neigh_ind,
        source=None,
        target=None,
        assume_sorted=None,
        *args,
        **kwargs,
    ) -> CSLS:
        """Fit the model using target, neigh_dist, and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (colums).
        neigh_ind: np.ndarray, shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source
            ignored
        target
            ignored
        assume_sorted: bool, default=True #noqa: DAR103
            Assume input matrices are sorted according to neigh_dist.
            If False, these are sorted here.
        *args
            Ignored
        **kwargs
            Ignored
        Returns
        -------
        CSLS
            Fitted CSLS
        Raises
        ------
        ValueError
            If self.k < 0
        TypeError
            If self.k not int
        """
        # Check equal number of rows and columns
        check_consistent_length(neigh_ind, neigh_dist)
        check_consistent_length(neigh_ind.T, neigh_dist.T)
        try:
            if self.k <= 0:
                raise ValueError(f"Expected k > 0. Got {self.k}")
        except TypeError:
            raise TypeError(f"Expected k: int > 0. Got {self.k}")

        # increment to include the k-th element in slicing
        k = self.k + 1

        if assume_sorted:
            self.r_dist_train_ = neigh_dist[:, :k]
            self.r_ind_train_ = neigh_ind[:, :k]
        else:
            kth = np.arange(self.k)
            mask = np.argpartition(neigh_dist, kth=kth)[:, :k]
            self.r_dist_train_ = np.take_along_axis(neigh_dist, mask, axis=1)
            self.r_ind_train_ = np.take_along_axis(neigh_ind, mask, axis=1)
        return self

    def transform(
        self,
        neigh_dist,
        neigh_ind,
        query,
        assume_sorted: bool = True,
        *args,
        **kwargs,
    ) -> (np.ndarray, np.ndarray):
        """Transform distance between test and training data with CSLS.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_query, n_neighbors)
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: np.ndarray, shape (n_query, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist
        query
            Ignored
        assume_sorted: bool
            ignored
        *args
            Ignored
        **kwargs
            Ignored

        Returns
        -------
        hub_reduced_dist, neigh_ind
            CSLS distances, and corresponding neighbor indices
        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(self, "r_dist_train_")

        n_test, n_indexed = neigh_dist.shape

        if n_indexed == 1:
            warnings.warn(
                "Cannot perform hubness reduction with a single neighbor per query. "
                "Skipping hubness reduction, and returning untransformed distances."
            )
            return neigh_dist, neigh_ind

        k = self.k

        # Find average distances to the k nearest neighbors
        if assume_sorted:
            r_dist_test = neigh_dist[:, :k]
        else:
            kth = np.arange(self.k)
            mask = np.argpartition(neigh_dist, kth=kth)[:, :k]
            r_dist_test = np.take_along_axis(neigh_dist, mask, axis=1)

        hub_reduced_dist = np.empty_like(neigh_dist)

        # Optionally show progress of local scaling loop
        disable_tqdm = not self.verbose
        range_n_test = tqdm(
            range(n_test),
            desc="CSLS",
            disable=disable_tqdm,
        )

        r_train = self.r_dist_train_.mean(axis=1)
        r_test = r_dist_test.mean(axis=1)
        for i in range_n_test:
            hub_reduced_dist[i, :] = (
                2 * neigh_dist[i] - r_test[i] - r_train[neigh_ind[i]]
            )
        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/

from __future__ import annotations

import warnings

import numpy as np
from scipy import stats
from sklearn.utils.validation import (
    check_array,
    check_consistent_length,
    check_is_fitted,
)
from tqdm.auto import tqdm

from .base import HubnessReduction


class MutualProximity(HubnessReduction):
    """Hubness reduction with Mutual Proximity.

    Uses the formula presented in [1]_.

    Parameters
    ----------
    method: 'normal' or 'empiric', default = 'normal'
        Model distance distribution with 'method'.
        - 'normal' or 'gaussi' model distance distributions with independent Gaussians (fast)
        - 'empiric' or 'exact' model distances with the empiric distributions (slow)
    verbose: int, default = 0
        If verbose > 0, show progress bar.
    References
    ----------
    .. [1] Schnitzer, D., Flexer, A., Schedl, M., & Widmer, G. (2012).
           Local and global scaling reduce hubs in space. The Journal of Machine
           Learning Research, 13(1), 2871–2902.
    """

    def __init__(self, method: str = "normal", verbose: int = 0, **kwargs):
        super().__init__(**kwargs)
        if method not in ["exact", "empiric", "normal", "gaussi"]:
            raise ValueError(
                f'Mutual proximity method "{method}" not recognized. Try "normal"'
                ' or "empiric".'
            )
        if method in ["exact", "empiric"]:
            self.method = "empiric"
        elif method in ["normal", "gaussi"]:
            self.method = "normal"
        self.verbose = verbose

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(method = {self.method}, verbose ="
            f" {self.verbose})"
        )

    def fit(
        self, neigh_dist, neigh_ind, source, target, assume_sorted=None
    ) -> MutualProximity:
        """Fit the model using neigh_dist and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (columns).
        neigh_ind: np.ndarray, shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source
            Ignored
        target
            Ignored
        assume_sorted
            Ignored
        Returns
        ------
        MutualProximity

        Raises
        ------
        ValueError
            If self.method is unknown
        """
        # Check equal number of rows and columns
        check_consistent_length(neigh_ind, neigh_dist)
        check_consistent_length(neigh_ind.T, neigh_dist.T)
        check_array(neigh_dist, force_all_finite=False)
        check_array(neigh_ind)

        self.n_train = neigh_dist.shape[0]

        if self.method not in ["normal", "empiric"]:
            raise ValueError(f"Internal: Invalid method {self.method}.")
        elif self.method == "empiric":
            self.neigh_dist_t_to_s_ = neigh_dist
            self.neigh_ind_t_to_s_ = neigh_ind
        elif self.method == "normal":
            self.mu_t_to_s_ = np.nanmean(neigh_dist, axis=1)
            self.sd_t_to_s_ = np.nanstd(neigh_dist, axis=1, ddof=0)
        return self

    def transform(self, neigh_dist, neigh_ind, query, assume_sorted=None):
        """Transform distance between test and training data with Mutual Proximity.

        Parameters
        ----------
        neigh_dist: np.ndarray
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: np.ndarray
            Neighbor indices corresponding to the values in neigh_dist
        query
            Ignored
        assume_sorted
            Ignored
        Returns
        -------
        hub_reduced_dist, neigh_ind
            Mutual Proximity distances, and corresponding neighbor indices
        Raises
        ------
        ValueError
            if self.method is unknown
        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(
            self,
            [
                "mu_t_to_s_",
                "sd_t_to_s_",
                "neigh_dist_t_to_s_",
                "neigh_ind_t_to_s_",
            ],
            all_or_any=any,
        )
        check_array(neigh_dist, force_all_finite="allow-nan")
        check_array(neigh_ind)

        n_test, n_indexed = neigh_dist.shape

        if n_indexed == 1:
            warnings.warn(
                "Cannot perform hubness reduction with a single neighbor per query. "
                "Skipping hubness reduction, and returning untransformed distances."
            )
            return neigh_dist, neigh_ind

        hub_reduced_dist = np.empty_like(neigh_dist)

        # Show progress in hubness reduction loop
        disable_tqdm = not self.verbose
        range_n_test = tqdm(
            range(n_test),
            desc=f"MP ({self.method})",
            disable=disable_tqdm,
        )

        # Calculate MP with independent Gaussians
        if self.method not in ["normal", "empiric"]:
            raise ValueError(f"Internal: Invalid method {self.method}.")
        elif self.method == "normal":
            mu_t_to_s = self.mu_t_to_s_
            sd_t_to_s_ = self.sd_t_to_s_
            for i in range_n_test:
                j_mom = neigh_ind[i]
                mu = np.nanmean(neigh_dist[i])
                sd = np.nanstd(neigh_dist[i], ddof=0)
                p1 = stats.norm.sf(neigh_dist[i, :], mu, sd)
                p2 = stats.norm.sf(
                    neigh_dist[i, :], mu_t_to_s[j_mom], sd_t_to_s_[j_mom]
                )
                hub_reduced_dist[i, :] = (1 - p1 * p2).ravel()
        # Calculate MP empiric (slow)
        elif self.method == "empiric":
            max_ind = max(self.neigh_ind_t_to_s_.max(), neigh_ind.max())
            for i in range_n_test:
                d_i = neigh_dist[i, :][np.newaxis, :]  # broadcasted afterwards
                d_j = np.zeros((d_i.size, n_indexed))
                for j in range(n_indexed):
                    tmp = np.zeros(max_ind + 1) + (
                        self.neigh_dist_t_to_s_[neigh_ind[i, j], -1] + 1e-6
                    )
                    tmp[
                        self.neigh_ind_t_to_s_[neigh_ind[i, j]]
                    ] = self.neigh_dist_t_to_s_[neigh_ind[i, j]]
                    d_j[j, :] = tmp[neigh_ind[i]]
                d = d_i.T
                hub_reduced_dist[i, :] = 1.0 - (
                    np.sum((d_i > d) & (d_j > d), axis=1) / n_indexed
                )

        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind



import sys

if sys.version_info[1] > 7:
    from importlib.metadata import version  # pragma: no cover
else:
    from importlib_metadata import version  # pragma: no cover

from kiez.kiez import Kiez

__version__ = version(__package__)

__all__ = ["Kiez"]



import warnings
from abc import ABC, abstractmethod

import numpy as np
from joblib import Parallel, delayed, effective_n_jobs
from sklearn.utils import gen_even_slices
from sklearn.utils.validation import check_is_fitted


class NNAlgorithm(ABC):
    """Base class for nearest neighbor algorithms"""

    valid_metrics = []

    def __init__(self, n_candidates, metric, n_jobs):
        self.n_candidates = n_candidates
        self.metric = metric
        self.n_jobs = n_jobs

    def _describe_source_target_fitted(self):
        if hasattr(self, "source_"):
            return (
                f" is fitted with: source.shape={self.source_.shape} and"
                f" target.shape={self.target_.shape}"
            )
        return " is unfitted"

    @property
    def valid_metrics(self):
        pass  # pragma: no cover

    @abstractmethod
    def _fit(self, data, is_source: bool):
        pass  # pragma: no cover

    def fit(self, source: np.ndarray, target: np.ndarray):
        """Indexes the given data using the underlying algorithm

        Parameters
        ----------
        source : matrix of shape (n_samples, n_features)
            embeddings of source entities
        target : matrix of shape (m_samples, n_features)
            embeddings of target entities

        Raises
        ------
        ValueError
            If source and target have a different number of features
        """
        if source.shape[1] != target.shape[1]:
            raise ValueError(
                "Expected source and target to have the same number of features, but"
                f" got source.shape: {source.shape} and target.shape: {target.shape}"
            )
        self.source_equals_target = np.array_equal(source, target)
        if self.source_equals_target:
            self.source_index = self._fit(source, True)
            self.target_index = self.source_index
        else:
            self.source_index = self._fit(source, True)
            self.target_index = self._fit(target, False)
        self.source_ = source
        self.target_ = target

    def _check_k_value(self, k, needed_space):
        if not np.issubdtype(type(k), np.integer):
            raise TypeError(f"k does not take {type(k)} value, enter integer value")
        if k <= 0:
            raise ValueError(f"Expected k > 0. Got {k}")
        if k > needed_space:
            warnings.warn(
                f"k={k} is larger than number of samples in indexed space.\n"
                + f"Setting to k={needed_space}"
            )
            return needed_space
        return k

    @abstractmethod
    def _kneighbors(self, query, k, index, return_distance, is_self_querying):
        pass  # pragma: no cover

    def kneighbors(self, query=None, k=None, s_to_t=True, return_distance=True):
        check_is_fitted(self, ["source_index", "target_index"])
        k = self.n_candidates if k is None else k
        is_self_querying = query is None and self.source_equals_target

        if s_to_t:
            query = self.source_ if query is None else query
            index = self.target_index
            needed_space = self.target_.shape[0]
        else:
            query = self.target_ if query is None else query
            index = self.source_index
            needed_space = self.source_.shape[0]
        k = self._check_k_value(k, needed_space)
        return self._kneighbors(
            k=k,
            query=query,
            index=index,
            return_distance=return_distance,
            is_self_querying=is_self_querying,
        )


class NNAlgorithmWithJoblib(NNAlgorithm):
    @abstractmethod
    def _kneighbors_part(self, k, query, index, return_distance, is_self_querying):
        pass  # pragma: no cover

    def _kneighbors(self, k, query, index, return_distance, is_self_querying):
        n_jobs = effective_n_jobs(self.n_jobs)
        # assume joblib>=0.12
        delayed_query = delayed(self._kneighbors_part)
        parallel_kwargs = {"prefer": "threads"}
        result = Parallel(n_jobs, **parallel_kwargs)(
            delayed_query(
                k=k,
                query=query[s],
                index=index,
                return_distance=return_distance,
                is_self_querying=is_self_querying,
            )
            for s in gen_even_slices(query.shape[0], n_jobs)
        )
        if return_distance:
            dist, neigh_ind = zip(*result)
            result = [
                np.atleast_2d(arr) for arr in [np.vstack(dist), np.vstack(neigh_ind)]
            ]
        else:
            result = np.atleast_2d(np.vstack(result))
        return result



from kiez.neighbors.approximate.hnsw import HNSW
from kiez.neighbors.approximate.nng import NNG
from kiez.neighbors.approximate.random_projection_trees import Annoy
from kiez.neighbors.exact.sklearn_nearest_neighbors import SklearnNN
from kiez.neighbors.neighbor_algorithm_base import NNAlgorithm

__all__ = ["HNSW", "NNG", "Annoy", "SklearnNN", "NNAlgorithm"]



# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness https://github.com/VarIr/scikit-hubness/
import sys


def available_ann_algorithms_on_current_platform():
    """Get approximate nearest neighbor algorithms available for the current platform/OS
    Currently, the algorithms are provided by the following libraries:
        * 'hnsw': nmslib
        * 'rptree': annoy
        * 'nng': NGT
    Returns
    -------
    algorithms: Tuple[str]
        A tuple of available algorithms
    """
    # Windows
    if sys.platform == "win32":
        algorithms = (
            "hnsw",
            "rptree",
        )
    # Mac and Linux
    elif sys.platform in ["linux", "darwin"]:
        algorithms = (
            "hnsw",
            "rptree",
            "nng",
        )
    # others undefined
    else:  # pragma: no cover
        algorithms = ()

    return algorithms



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kiez',
 'kiez.analysis',
 'kiez.evaluate',
 'kiez.hubness_reduction',
 'kiez.io',
 'kiez.neighbors',
 'kiez.neighbors.approximate',
 'kiez.neighbors.exact',
 'kiez.utils']

package_data = \
{'': ['*']}

install_requires = \
['annoy>=1.17.0,<2.0.0',
 'importlib-metadata',
 'joblib>=1.0.0,<2.0.0',
 'ngt>=1.12.2,<2.0.0',
 'nmslib>=2.1.1,<3.0.0',
 'numpy>=1.20.0,<2.0.0',
 'pandas>=1.2.1,<2.0.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'scipy>=1.6.0,<2.0.0',
 'tqdm>=4.56.0,<5.0.0']

extras_require = \
{'docs': ['sphinx>=4.0.2,<5.0.0', 'insegel>=1.1.0,<2.0.0']}

setup_kwargs = {
    'name': 'kiez',
    'version': '0.2.1',
    'description': 'Hubness reduced nearest neighbor search for entity alignment with knowledge graph embeddings',
    'long_description': '<p align="center">\n<img src="https://github.com/dobraczka/kiez/raw/develop/docs/kiezlogo.png" alt="kiez logo", width=200/>\n</p>\n\n<h2 align="center"> kiez</h2>\n\n<p align="center">\n<a href="https://github.com/dobraczka/kiez/actions/workflows/main.yml"><img alt="Actions Status" src="https://github.com/dobraczka/kiez/actions/workflows/main.yml/badge.svg?branch=main"></a>\n<a><img alt="Test coverage" src="https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/dobraczka/7c57dda3b055c972a06f0f076df46196/raw/test.json"></a>\n<a href="https://github.com/dobraczka/kiez/blob/main/LICENSE"><img alt="License BSD3 - Clause" src="https://img.shields.io/badge/license-BSD--3--Clause-blue"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\nA Python library for hubness reduced nearest neighbor search for the task of entity alignment with knowledge graph embeddings. The term kiez is a [german word](https://en.wikipedia.org/wiki/Kiez) that refers to a city neighborhood.\n\n## Hubness Reduction\nHubness is a phenomenon that arises in high-dimensional data and describes the fact that a couple of entities are nearest neighbors (NN) of many other entities, while a lot of entities are NN to no one.\nFor entity alignment with knowledge graph embeddings we rely on NN search. Hubness therefore is detrimental to our matching results.\nThis library is intended to make hubness reduction techniques available to data integration projects that rely on knowledge graph embeddings in their alignment process. Furthermore kiez optionally incorporates several approximate nearest neighbor (ANN) libraries, to pair the speed advantage of approximate neighbor search with increased accuracy of hubness reduction.\n\n## Installation\nTo install the base package without ANN libraries\n``` bash\npip install kiez\n```\n\n## Usage\nSimple nearest neighbor search for source entities in target space:\n``` python\nfrom kiez import Kiez\nimport numpy as np\n# create example data\nrng = np.random.RandomState(0)\nsource = rng.rand(100,50)\ntarget = rng.rand(100,50)\n# fit and get neighbors\nk_inst = Kiez()\nk_inst.fit(source, target)\nnn_dist, nn_ind = k_inst.kneighbors()\n```\nUsing ANN libraries and hubness reduction methods:\n``` python\nfrom kiez import Kiez\nimport numpy as np\n# create example data\nrng = np.random.RandomState(0)\nsource = rng.rand(100,50)\ntarget = rng.rand(100,50)\n# prepare algorithm and hubness reduction\nfrom kiez.neighbors import HNSW\nhnsw = HNSW(n_candidates=10)\nfrom kiez.hubness_reduction import CSLS\nhr = CSLS()\n# fit and get neighbors\nk_inst = Kiez(n_neighbors=5, algorithm=hnsw, hubness=hr)\nk_inst.fit(source, target)\nnn_dist, nn_ind = k_inst.kneighbors()\n```\n\nLicense\n-------\n`kiez` is licensed under the terms of the BSD-3-Clause [license](LICENSE.txt).\nSeveral files were modified from [`scikit-hubness`](https://github.com/VarIr/scikit-hubness),\ndistributed under the same [license](external/SCIKIT_HUBNESS_LICENSE.txt).\nThe respective files contain the following tag instead of the full license text.\n\n        SPDX-License-Identifier: BSD-3-Clause\n\nThis enables machine processing of license information based on the SPDX\nLicense Identifiers that are here available: https://spdx.org/licenses/\n',
    'author': 'Daniel Obraczka',
    'author_email': 'obraczka@informatik.uni-leipzig.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dobraczka/kiez',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/

from __future__ import annotations

import warnings

import numpy as np
from sklearn.utils.validation import check_consistent_length, check_is_fitted
from tqdm.auto import tqdm

from .base import HubnessReduction


class LocalScaling(HubnessReduction):
    """Hubness reduction with Local Scaling.

    Uses the formula presented in [1]_.

    Parameters
    ----------
    k: int, default = 5
        Number of neighbors to consider for the rescaling
    method: 'standard' or 'nicdm', default = 'standard'
        Perform local scaling with the specified variant:
        - 'standard' or 'ls' rescale distances using the distance to the k-th neighbor
        - 'nicdm' rescales distances using a statistic over distances to k neighbors
    verbose: int, default = 0
        If verbose > 0, show progress bar.
    References
    ----------
    .. [1] Schnitzer, D., Flexer, A., Schedl, M., & Widmer, G. (2012).
           Local and global scaling reduce hubs in space. The Journal of Machine
           Learning Research, 13(1), 2871–2902.
    """

    def __init__(
        self, k: int = 5, method: str = "standard", verbose: int = 0, **kwargs
    ):
        super().__init__(**kwargs)
        self.k = k
        self.method = method
        self.verbose = verbose

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(k={self.k}, method = {self.method}, verbose ="
            f" {self.verbose})"
        )

    def fit(
        self,
        neigh_dist,
        neigh_ind,
        source,
        target,
        assume_sorted: bool = True,
        *args,
        **kwargs,
    ) -> LocalScaling:
        """Fit the model using neigh_dist and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (colums).
        neigh_ind: np.ndarray, shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source
            Ignored
        target
            Ignored
        assume_sorted: bool, default = True #noqa: DAR103
            Assume input matrices are sorted according to neigh_dist.
            If False, these are sorted here.
        *args
            Ignored
        **kwargs
            Ignored
        Returns
        -------
        LocalScaling
            Fitted LocalScaling
        """
        # Check equal number of rows and columns
        check_consistent_length(neigh_ind, neigh_dist)
        check_consistent_length(neigh_ind.T, neigh_dist.T)

        # increment to include the k-th element in slicing
        k = self.k + 1

        # Find distances to the k-th neighbor (standard LS) or the k neighbors (NICDM)
        if assume_sorted:
            self.r_dist_t_to_s_ = neigh_dist[:, :k]
            self.r_ind_t_to_s_ = neigh_ind[:, :k]
        else:
            kth = np.arange(self.k)
            mask = np.argpartition(neigh_dist, kth=kth)[:, :k]
            self.r_dist_t_to_s_ = np.take_along_axis(neigh_dist, mask, axis=1)
            self.r_ind_t_to_s_ = np.take_along_axis(neigh_ind, mask, axis=1)
        return self

    def transform(
        self,
        neigh_dist,
        neigh_ind,
        source=None,
        assume_sorted: bool = True,
    ) -> (np.ndarray, np.ndarray):
        """Transform distance between test and training data with Mutual Proximity.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_query, n_neighbors)
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: np.ndarray, shape (n_query, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist
        source
            Ignored
        assume_sorted: bool, default = True #noqa: DAR103
            Assume input matrices are sorted according to neigh_dist.
            If False, these are partitioned here.
            NOTE: The returned matrices are never sorted.
        Returns
        -------
        hub_reduced_dist, neigh_ind
            Local scaling distances, and corresponding neighbor indices
        Raises
        ------
        ValueError
            If wrong self.method was supplied
        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(self, "r_dist_t_to_s_")

        n_test, n_indexed = neigh_dist.shape

        if n_indexed == 1:
            warnings.warn(
                "Cannot perform hubness reduction with a single neighbor per query. "
                "Skipping hubness reduction, and returning untransformed distances."
            )
            return neigh_dist, neigh_ind

        # increment to include the k-th element in slicing
        k = self.k + 1

        # Find distances to the k-th neighbor (standard LS) or the k neighbors (NICDM)
        if assume_sorted:
            r_dist_s_to_t = neigh_dist[:, :k]
        else:
            kth = np.arange(self.k)
            mask = np.argpartition(neigh_dist, kth=kth)[:, :k]
            r_dist_s_to_t = np.take_along_axis(neigh_dist, mask, axis=1)

        # Calculate LS or NICDM
        hub_reduced_dist = np.empty_like(neigh_dist)

        # Optionally show progress of local scaling loop
        disable_tqdm = not self.verbose
        range_n_test = tqdm(
            range(n_test),
            desc=f"LS {self.method}",
            disable=disable_tqdm,
        )

        # Perform standard local scaling...
        if self.method not in ["ls", "standard", "nicdm"]:
            raise ValueError(
                f"Internal: Invalid method {self.method}. Try 'ls' or 'nicdm'."
            )
        elif self.method in ["ls", "standard"]:
            r_t_to_s = self.r_dist_t_to_s_[:, -1]
            r_s_to_t = r_dist_s_to_t[:, -1]
            for i in range_n_test:
                hub_reduced_dist[i, :] = 1.0 - np.exp(
                    -1 * neigh_dist[i] ** 2 / (r_s_to_t[i] * r_t_to_s[neigh_ind[i]])
                )
        # ...or use non-iterative contextual dissimilarity measure
        elif self.method == "nicdm":
            r_t_to_s = self.r_dist_t_to_s_.mean(axis=1)
            r_s_to_t = r_dist_s_to_t.mean(axis=1)
            for i in range_n_test:
                hub_reduced_dist[i, :] = neigh_dist[i] / np.sqrt(
                    (r_s_to_t[i] * r_t_to_s[neigh_ind[i]])
                )

        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind
